from tqdm import tqdm 

import numpy as np
import pandas as pd
import dolphindb as ddb

from DDBBase import DDBBase
from DDBLoader import DDBPITLoader


class DDBExpression(DDBBase):
    """
    调用事先定义好的DolphinDB的Function或者FunctionView来进行表达式计算。

    定义DolphinDB中的Function或者FunctionView的具体逻辑在`DDBExpreCreator`中。
    """

    def pit_at_date(self, table_name, date, report_period_list):
        """
        获得任何时间点上，能看到的某个财报最新可见数据。

        实现方式：
            通过调用事先在DolpinDB中创建的Function View。

        Args:
            table_name: 财务报表名字，包含是否为调整报表的后缀名（非stem名）
            date: 观察日期，使用DolphinDB的日期格式，即YYYY.MM.DD，无需加引号
            report_period_list: 报告期列表，列表中每一个日期为一个DolphinDb的日期格式

        Return:
            DataFrame: 在观察日期`date`日能看到的，所有公司最近能看到的，报告期在`report_period_list`中的，财务报告表`table_name`的截面数据。
        """
        df_list = []
        for i in range(DDBPITLoader.num_code_partition):
            _df = self.ddb_sess.run("""
                select * from pit_at_date("{table_name}", {date}, [{report_period_list}], {code_partition_id});
            """.format(
                table_name = table_name,
                date = date,
                report_period_list = ','.join(report_period_list),
                code_parition_id = i
            ))
            _df.set_index(['code', 'report_period', 'appear_at_date'], inplace=True)
            df_list.append(_df)
        return pd.concat(df_list, axis=1)


    def pit_col_at_date(self, table_name, col_name, date, report_period_list):
        """
        获得任何时间点上，能看到的某个财报最新可见数据。

        实现方式：
            通过调用事先在DolpinDB中创建的Function View。

        Args:
            table_name: 财务报表名字，包含是否为调整报表的后缀名（非stem名）
            col_name: 列名，一般为具体财务科目名字。
            date: 观察日期，使用DolphinDB的日期格式，即YYYY.MM.DD，无需加引号
            report_period_list: 报告期列表，列表中每一个日期为一个DolphinDb的日期格式

        Return:
            DataFrame: 在观察日期`date`日能看到的，所有公司最近能看到的，报告期在`report_period_list`中的，财务报告表`table_name`中``col_name`指标的截面数据。
 
        """
        df_list = []
        for i in range(DDBPITLoader.num_code_partition):
            _df = self.ddb_sess.run("""
                select * from pit_col_at_date("{table_name}", "{col_name}", {date}, [{report_period_list}], {code_partition_id});
            """.format(
                table_name = table_name,
                col_name = col_name,
                date = date,
                report_period_list = ','.join(report_period_list),
                code_parition_id = i
            ))
            _df.set_index(['code', 'report_period', 'appear_at_date'], inplace=True)
            df_list.append(_df)
        return pd.concat(df_list, axis=1)


    def yoy(self, dataframe, col_list):
        self.ddb_sess.upload({'table_yoy' : dataframe })
        self.ddb_sess.run("""
            yoy({table_name}, [{col_list}]);
        """.format(
            table_name = 'table_yoy',
            col_list = ','.join([f"'{col_name}'" for col_name in col_list])
        ))


    def single_quarter(self, dataframe, col_list):
        self.ddb_sess.upload({'table_single_quarter' : dataframe})
        self.ddb_sess.run("""
            single_quarter({table_name}, [{col_list}])
        """.format(
            table_name = table_name,
            col_list = ','.join([f"'{col_name}'" for col_name in col_list])
        ))



class DDBExpreCreator(DDBBase):
    """
    定义DolphinDB中的各种基础操作，因此这个类中的大部分实质性代码都是DolphinDB自己的语言，而不是Python语言编写的。

    `Expre`为`Expression`的简写。
    """
    pit_dbpath = "dfs://pit_stock_ts"
    pit_dbname = "db_pit_stock"
    
    daily_dbpath = "dfs://daily_stock_ts"
    daily_dbname = "db_daily_stock"


    def create_function_views(self):
        self._func_view_pit_at_date()
        self._func_view_pit_col_at_date()


    def create_functions(self):
        self._fun_yoy();
        self._func_single_quarter()


    def _func_single_quarter(self):
        code = """
            /**
			 * Args:
			 *     table: DolphinDB的表，可以是一个内存表，也可以是一个分区表。
             *     col_list: 需要进行单季转换的字段名列表。
			 */ 
			def single_quarter(table_name, col_list) {
			
				sel_col = array(ANY, 4 + size(col_list));
				sel_col[0] = sqlCol('code');
				sel_col[2] = sqlColAlias(<year(report_period)>, 'year');
				
				for (i in 0..(size(col_list) - 1)) {
					sel_col[i + 4] = sqlCol(col_list[i]);
				}

				// 当季累计数据，作为下一季基准
				sel_col[1] = sqlColAlias(sqlCol('report_period'), 'base_report_period');
				sel_col[3] = sqlColAlias(<quarterOfYear(report_period) + 1>, 'quarter_of_year');
				tbl_quarter_accum_base = sql(
						select = sel_col,
						from = table_name,
						where = <quarterOfYear(report_period) < 4>
					).eval();

				// 从第二季开始，需要去匹配前一季累计基数
				sel_col[1] = sqlColAlias(sqlCol('report_period'), 'current_report_period');
				sel_col[3] = sqlColAlias(<quarterOfYear(report_period)>, 'quarter_of_year');
				tbl_quarter_accum = sql(
							select = sel_col,
							from = table_name,
							where = <quarterOfYear(report_period) > 1>
						).eval();

				// 单季流量，把非第一季的季报都要扣除前一季度的基数
				sel_col[1] = sqlColAlias(sqlCol('current_report_period'), 'report_period');
				sel_col[2] = sqlCol('year');
				sel_col[3] = sqlCol('quarter_of_year');
				for (i in 0..(size(col_list) - 1)) {
					sel_col[i + 4] = sqlColAlias(
						expr(
							sqlCol('tbl_quarter_accum_' + col_list[i]),
							-,
							sqlCol(col_list[i])
						)
					, col_list[i] + '_flux');
				}
				from_obj = ej(
							tbl_quarter_accum_base, 
							tbl_quarter_accum, 
							`code`year`quarter_of_year, 
							`code`year`quarter_of_year);
				tbl_quarter_flux = sql(
						select = sel_col, 
						from = from_obj
					).eval();

				// 每年第一个季度
				sel_col[1] = sqlCol('report_period');
				sel_col[2] = sqlColAlias(<year(report_period)>, 'year');
				sel_col[3] = sqlColAlias(<quarterOfYear(report_period)>, 'quarter_of_year');
				for (i in 0..(size(col_list) - 1)) {
					sel_col[i + 4] = sqlColAlias(sqlCol(col_list[i]), col_list[i] + '_flux');
				}
				tbl_quarter1 = sql(
						select = sel_col,
						from = table_name,
						where = <quarterOfYear(report_period) == 1>
					).eval();
				
				// 再拼接回第一季度（无需扣除基数的数据）
				tbl_quarter_flux = unionAll(
						tbl_quarter1, 
						tbl_quarter_flux
					);
					
				return tbl_quarter_flux;
			}
        """
		self.ddb_sess.run(code)


    def _func_yoy(self):
        code = """
			/**
			 * Args:
			 *     table: DolphinDB的表，可以是一个内存表，也可以是一个分区表。
             *     col_list: 需要进行同比转换的字段名。
			 */
			def yoy(table_name, col_list) {
				sel_col = array(ANY, 4 + size(col_list));
				sel_col[0] = sqlCol('code');
				sel_col[3] = sqlColAlias(<quarterOfYear(report_period)>, 'quarter_of_year');
				
				// 上一年数据
				for (i in 0..(size(col_list) - 1)) {
					sel_col[i + 4] = sqlColAlias(sqlCol(col_list[i]), col_list[i] + '_base');
				}
				sel_col[1] = sqlColAlias(sqlCol('report_period'), 'report_period_last_year');
				sel_col[2] = sqlColAlias(<year(report_period) + 1>, 'year');
				tbl_last_year = sql(
						select = sel_col,
						from = table_name
					).eval();

				// 本年度数据
				for (i in 0..(size(col_list) - 1)) {
					sel_col[i + 4] = sqlCol(col_list[i]);
				}
				sel_col[1] = sqlColAlias(sqlCol('report_period'), 'report_period_current_year');
				sel_col[2] = sqlColAlias(<year(report_period)>, 'year');
				tbl_this_year = sql(
						select = sel_col,
						from = table_name
					).eval();
				
				// 计算同比增长率
				sel_col = array(ANY, 4 + (3 * size(col_list)));

				sel_col[0] = sqlCol('code');
				sel_col[1] = sqlColAlias(sqlCol('report_period_current_year'), 'report_period');
				sel_col[2] = sqlCol('year');
				sel_col[3] = sqlCol('quarter_of_year');
				
				for (i in 0..(size(col_list) - 1)) {
					sel_col[3 * i + 4] = sqlCol(col_list[i] + '_base');
					sel_col[3 * i + 5] = sqlCol(col_list[i]);
					sel_col[3 * i + 6] = sqlColAlias(
						expr(expr(sqlCol(col_list[i]), -, sqlCol(col_list[i] + '_base')), /, expr(abs, sqlCol(col_list[i] + '_base'))), 
						col_list[i] + '_yoy');
				}
				from_obj = ej(
						tbl_last_year, 
						tbl_this_year, 
						`code`year`quarter_of_year, 
						`code`year`quarter_of_year);
						
				return sql(
						select = sel_col,
						from = from_obj).eval();
			} 
        """
        self.ddb_sess.run(code)


    def _func_view_pit_at_date(self):
        """
        创建`pit_at_date`，获得任何时间点上，能看到的某个财报最新可见数据。
        """
        self.ddb_sess.run("""
            dropFunctionView('pit_at_date');
        """)

        self.ddb_sess.run("""
			/**
			 * Example of use:
			 *     pit_at_date("is_common_ori", "TOT_OPER_REV", 2021.03.14, [2019.12.31, 2019.09.30, 2019.06.30], 0)
			 */
            def pit_at_date(table_name, date, report_period_list, code_partition_id) {
                source_table = loadTable("{pit_dbpath}", table_name);
                
                m_nDate = take(date, size(report_period_list));
                report_period = report_period_list;
                
                query_table = table(report_period, m_nDate);
                query_table_exp = select * from cj(
                        query_table, 
                        select code 
                            from source_table 
                            where partition(code, code_partition_id) 
                            group by code map
                    );
                
                source_table_part = select source_table.* from ej(
                        source_table, query_table_exp, `code`report_period
                    ) where partition(code, code_partition_id);
                
                return select source_table_part.* from aj(
                        query_table_exp, 
                        source_table_part, 
                        `code`report_period`m_nDate, 
                        `code`report_period`appear_at_date
                    ) where not isNull(source_table_part.code);
            }

            addFunctionView(pit_at_date);
        """.format(
            pit_dbpath = pit_dbpath
        ))

    
    def _func_view_pit_col_at_date(self):
        """
        创建`pit_col_at_date`，获得任何日期点上，能看到的财报中某一个指标的最新可见数据。
        与`pit_at_date`的区别在于，`pit_col_at_date`返回返回一个序列，而`pit_at_date`返回一整张表。
        """
        # 先把原有的View Function删除
        self.ddb_sess.run("""
            dropFunctionView('pit_col_at_date');
        """)

        self.ddb_sess.run("""
            /**
			 * Example of use:
			 *     pit_col_at_date("is_common_ori", "TOT_OPER_REV", 2021.03.14, [2019.12.31, 2019.09.30, 2019.06.30], 0)
			 */
			def pit_col_at_date(table_name, col_name, date, report_period_list, code_partition_id) {
                source_table = loadTable("{pit_dbpath}", table_name);
                
                m_nDate = take(date, size(report_period_list));
                report_period = report_period_list;
                
                query_table = table(report_period, m_nDate);
                query_table_exp = select * from cj(
                        query_table, 
                        select code 
                            from source_table 
                            where partition(code, code_partition_id) 
                            group by code map
                    );
                
                col_list = sqlCol(['code', 'report_period', 'appear_at_date', col_name]);
                from_tbl = <ej(source_table, query_table_exp, `code`report_period)>;
                where_conditions = [<partition(code, code_partition_id)>];
                source_table_part = sql(select=col_list, from=from_tbl, where=where_conditions).eval();
                
                return select source_table_part.* from aj(
                        query_table_exp, 
                        source_table_part, 
                        `code`report_period`m_nDate, 
                        `code`report_period`appear_at_date
                    ) where not isNull(source_table_part.code)
            }

            addFunctionView(pit_col_at_date);
        """.format(
            pit_dbpath = self.pit_dbpath
        ))


