import pickle
import functools
import abc
import warnings

from pprint import pprint
from pathlib import Path
from tqdm import tqdm
#from tqdm.contrib.concurrent import process_map
from multiprocessing import Pool

import numpy as np
import pandas as pd
from pandas.core.common import SettingWithCopyWarning
warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

import dolphindb as ddb
import dolphindb.settings as keys

import sqlalchemy as sa

import ProtoBuffEntitys


class DDBDailyLoader(DDBLoader):

    ddb_path = "dfs://daily_stock_ts"
    ddb_dbname = "db_daily_stock_ts"

    daily_kline_cols = [
        'code', 'm_nDate', 
        'open', 'high', 'low', 'close', 'vol',
        'amount', 'cjbs', 'yclose', 
        'PctChg', 'IsZt', 'IsDt', 'IsST', 'IsGoDelist',
        'FloatShares', 'MarketValues',
        'factor'
    ]

    daily_kline_col_types = [
        'SYMBOL', 'DATE',
        'DOUBLE', 'DOUBLE', 'DOUBLE', 'DOUBLE', 'DOUBLE', 
        'DOUBLE', 'INT', 'DOUBLE',
        'DOUBLE', 'INT', 'INT', 'INT', 'INT',
        'DOUBLE', 'DOUBLE',
        'DOUBLE'
    ]


    def create_ddb_database(self):
        # TODO: daily数据库已经在DDBDailyFactor中被创建了
        #       后续可以迁移过来，不过现在就暂时先不管了
        pass


    def load_ddb_database(self):
        self.ddb_sess.run("""
            {dbName} = database(directory='{dbPath}')
        """.format(
            dbName = self.ddb_dbname,
            dbPath = self.ddb_path
        ))
        print('Did load database from', self.ddb_path)


    def create_ddb_partition_table(self, memory_table_name, partition_table_name):
        # TODO: 现在只做一个日频行情数据表，今后可能考虑把基本面数据也迁移过来

        # 由于日频行情数据的表结构相对简单，所以直接把表结构写在这里代码里即可
        # 搬迁数据的时候需要考虑按照逐个股票来搬迁，以免造成对内存的巨大压力
        self.ddb_sess.run("""
            // 确保删除原表
            if (existsTable("{ddb_daily_path}", "{partition_table_name}")) {{
                dropTable({ddb_daily_dbname}, "{partition_table_name}");
            }}

            // 然后根据内存表的结构，创建持久化的分区表
            {partition_table_name} = {ddb_daily_dbname}.createPartitionedTable(
                table = {memory_table_name},
                tableName = "{partition_table_name}",
                partitionColumns = `code,
                sortColumns = `code`m_nDate,
                compressMethods = {{m_nDate:"delta"}}
            );
        """.format(
            ddb_daily_path = self.ddb_path,
            ddb_daily_dbname = self.ddb_dbname,
            memory_table_name = memory_table_name,
            partition_table_name = partition_table_name,
        ))
    

    def create_ddb_memory_table(self, memory_table_name, capacity):
        self.ddb_sess.run("""
            // 先创建一个空的内存表用来表征结构，如果无需插入数据，capacity可以设为10
            {memory_table_name} = table({capacity}:0, {col_names}, [{col_types}]);
        """.format(
            memory_table_name = memory_table_name,
            capacity = capacity,
            col_names =  '`' + '`'.join(self.daily_kline_cols),
            col_types = ', '.join(self.daily_kline_col_types)
        ))


    def dump_daily_kline_to_ddb(self):
        # 先创建一个分区表，然后再逐个股票的数据插入
        # 1. 需要额外控制在插入第一个股票数据的时候创建分区表比较麻烦
        # 2. python程序中的dataframe直接上传到dolphindb内存表，不需要考虑内存表字段类型，分区表中设置好即可

        memory_table_name = 'daily_kline_mem'
        partition_table_name = 'daily_kline'

        self.create_ddb_memory_table(memory_table_name, 10)
        print('Did create ddb memory table.')
        pprint(self.ddb_sess.run(f"schema({memory_table_name})"))
        self.create_ddb_partition_table(memory_table_name, partition_table_name)
        print('Did create ddb partition table.')
        pprint(self.ddb_sess.run(f"schema({partition_table_name})"))

        with self.mssql_engine.connect() as conn:
            stat = "select distinct [StockID] from [StockDaily].dbo.[DailyKLine]"
            stock_id_list = list(conn.execute(stat).fetchall())

            with tqdm(stock_id_list) as pbar:
                for (stock_id,) in pbar:
                    pbar.set_description(f"Will work on {stock_id}") 
                    #pbar.set_description(f"Will fetch all data of {stock_id} from SQL Server")
                    stat = """
                        select * from [StockDaily].dbo.[DailyKLine]
                        where StockID='{stock_id}'
                    """.format(
                        stock_id = stock_id
                    )
                    row_list = list(conn.execute(stat).fetchall())
                    num_rows = len(row_list)

                    #pbar.set_description(f"Will work on dumping job on {stock_id} of len {num_rows}")
                    df = pd.DataFrame(row_list)
                    df['date'] = DDBLoader.make_date(df['date'])
                    df['StockID'] = DDBLoader.tscode_to_windcode(df['StockID'])
                    self.ddb_sess.upload({memory_table_name : df})
                    #print('Did upload dataframe to ddb.')
                    #pprint(self.ddb_sess.run(f"schema({memory_table_name})"))
                    #break
                    self.ddb_sess.run(f"{partition_table_name}.tableInsert({memory_table_name})")


def main():

	# TODO: 
	# 可以使用`Fire`库，对函数调用再做一次封装，就可以避免每次运行不同参数时候需要修改内部多处的代码。

    # 日频行情数据
    loader = DDBDailyLoader()
    loader.load_ddb_database()
    #loader.dump_daily_kline_to_ddb()

    
if __name__ == '__main__':
    main() 

